"""Core compute functions, for polar, flux, and cartesian coordinates."""

from desc.backend import jnp

from .data_index import register_compute_fun


@register_compute_fun(
    name="0",
    label="0",
    units="~",
    units_long="None",
    description="Zeros",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
        "desc.geometry.core.Curve",
    ],
    aliases=["rho_t", "rho_z", "theta_r", "theta_z", "zeta_r", "zeta_t"],
)
def _0(params, transforms, profiles, data, **kwargs):
    data["0"] = jnp.zeros(transforms["grid"].num_nodes)
    return data


@register_compute_fun(
    name="1",
    label="1",
    units="~",
    units_long="None",
    description="Ones",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
        "desc.geometry.core.Curve",
    ],
    aliases=["rho_r", "theta_t", "zeta_z"],
)
def _1(params, transforms, profiles, data, **kwargs):
    data["1"] = jnp.ones(transforms["grid"].num_nodes)
    return data


@register_compute_fun(
    name="R",
    label="R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R(params, transforms, profiles, data, **kwargs):
    data["R"] = transforms["R"].transform(params["R_lmn"], 0, 0, 0)
    return data


@register_compute_fun(
    name="R_r",
    label="\\partial_{\\rho} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, first radial derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_r(params, transforms, profiles, data, **kwargs):
    data["R_r"] = transforms["R"].transform(params["R_lmn"], 1, 0, 0)
    return data


@register_compute_fun(
    name="R_rr",
    label="\\partial_{\\rho \\rho} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second radial derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rr(params, transforms, profiles, data, **kwargs):
    data["R_rr"] = transforms["R"].transform(params["R_lmn"], 2, 0, 0)
    return data


@register_compute_fun(
    name="R_rrr",
    label="\\partial_{\\rho \\rho \\rho} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third radial derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[3, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrr(params, transforms, profiles, data, **kwargs):
    data["R_rrr"] = transforms["R"].transform(params["R_lmn"], 3, 0, 0)
    return data


@register_compute_fun(
    name="R_rrrr",
    label="\\partial_{\\rho \\rho \\rho \\rho} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth radial derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[4, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrrr(params, transforms, profiles, data, **kwargs):
    data["R_rrrr"] = transforms["R"].transform(params["R_lmn"], 4, 0, 0)
    return data


@register_compute_fun(
    name="R_rrrt",
    label="\\partial_{\\rho \\rho \\rho \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt"
    " radial coordinate thrice and poloidal once",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[3, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrrt(params, transforms, profiles, data, **kwargs):
    data["R_rrrt"] = transforms["R"].transform(params["R_lmn"], 3, 1, 0)
    return data


@register_compute_fun(
    name="R_rrrz",
    label="\\partial_{\\rho \\rho \\rho \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt"
    " radial coordinate thrice and toroidal once",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[3, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rrrz(params, transforms, profiles, data, **kwargs):
    data["R_rrrz"] = transforms["R"].transform(params["R_lmn"], 3, 0, 1)
    return data


@register_compute_fun(
    name="R_rrt",
    label="\\partial_{\\rho \\rho \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative, wrt radius twice "
    "and poloidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrt(params, transforms, profiles, data, **kwargs):
    data["R_rrt"] = transforms["R"].transform(params["R_lmn"], 2, 1, 0)
    return data


@register_compute_fun(
    name="R_rrtt",
    label="\\partial_{\\rho \\rho \\theta \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative, wrt radius twice "
    "and poloidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrtt(params, transforms, profiles, data, **kwargs):
    data["R_rrtt"] = transforms["R"].transform(params["R_lmn"], 2, 2, 0)
    return data


@register_compute_fun(
    name="R_rrtz",
    label="\\partial_{\\rho \\rho \\theta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt radius twice,"
    " poloidal angle, and toroidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rrtz(params, transforms, profiles, data, **kwargs):
    data["R_rrtz"] = transforms["R"].transform(params["R_lmn"], 2, 1, 1)
    return data


@register_compute_fun(
    name="R_rrz",
    label="\\partial_{\\rho \\rho \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative, wrt radius twice "
    "and toroidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rrz(params, transforms, profiles, data, **kwargs):
    data["R_rrz"] = transforms["R"].transform(params["R_lmn"], 2, 0, 1)
    return data


@register_compute_fun(
    name="R_rrzz",
    label="\\partial_{\\rho \\rho \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative, wrt radius twice "
    "and toroidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[2, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rrzz(params, transforms, profiles, data, **kwargs):
    data["R_rrzz"] = transforms["R"].transform(params["R_lmn"], 2, 0, 2)
    return data


@register_compute_fun(
    name="R_rt",
    label="\\partial_{\\rho \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second derivative wrt radius "
    "and poloidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rt(params, transforms, profiles, data, **kwargs):
    data["R_rt"] = transforms["R"].transform(params["R_lmn"], 1, 1, 0)
    return data


@register_compute_fun(
    name="R_rtt",
    label="\\partial_{\\rho \\theta \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative wrt radius and "
    "poloidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rtt(params, transforms, profiles, data, **kwargs):
    data["R_rtt"] = transforms["R"].transform(params["R_lmn"], 1, 2, 0)
    return data


@register_compute_fun(
    name="R_rttt",
    label="\\partial_{\\rho \\theta \\theta \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt radius and "
    "poloidal angle thrice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 3, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _R_rttt(params, transforms, profiles, data, **kwargs):
    data["R_rttt"] = transforms["R"].transform(params["R_lmn"], 1, 3, 0)
    return data


@register_compute_fun(
    name="R_rttz",
    label="\\partial_{\\rho \\theta \\theta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt radius once, "
    "poloidal angle twice, and toroidal angle once",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 2, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rttz(params, transforms, profiles, data, **kwargs):
    data["R_rttz"] = transforms["R"].transform(params["R_lmn"], 1, 2, 1)
    return data


@register_compute_fun(
    name="R_rtz",
    label="\\partial_{\\rho \\theta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative wrt radius, poloidal "
    "angle, and toroidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rtz(params, transforms, profiles, data, **kwargs):
    data["R_rtz"] = transforms["R"].transform(params["R_lmn"], 1, 1, 1)
    return data


@register_compute_fun(
    name="R_rtzz",
    label="\\partial_{\\rho \\theta \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt radius, poloidal "
    "angle, and toroidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 1, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rtzz(params, transforms, profiles, data, **kwargs):
    data["R_rtzz"] = transforms["R"].transform(params["R_lmn"], 1, 1, 2)
    return data


@register_compute_fun(
    name="R_rz",
    label="\\partial_{\\rho \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second derivative wrt radius "
    "and toroidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rz(params, transforms, profiles, data, **kwargs):
    data["R_rz"] = transforms["R"].transform(params["R_lmn"], 1, 0, 1)
    return data


@register_compute_fun(
    name="R_rzz",
    label="\\partial_{\\rho \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative wrt radius and "
    "toroidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rzz(params, transforms, profiles, data, **kwargs):
    data["R_rzz"] = transforms["R"].transform(params["R_lmn"], 1, 0, 2)
    return data


@register_compute_fun(
    name="R_rzzz",
    label="\\partial_{\\rho \\zeta \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, fourth derivative wrt radius and "
    "toroidal angle thrice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[1, 0, 3]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _R_rzzz(params, transforms, profiles, data, **kwargs):
    data["R_rzzz"] = transforms["R"].transform(params["R_lmn"], 1, 0, 3)
    return data


@register_compute_fun(
    name="R_t",
    label="\\partial_{\\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, first poloidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _R_t(params, transforms, profiles, data, **kwargs):
    data["R_t"] = transforms["R"].transform(params["R_lmn"], 0, 1, 0)
    return data


@register_compute_fun(
    name="R_tt",
    label="\\partial_{\\theta \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second poloidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _R_tt(params, transforms, profiles, data, **kwargs):
    data["R_tt"] = transforms["R"].transform(params["R_lmn"], 0, 2, 0)
    return data


@register_compute_fun(
    name="R_ttt",
    label="\\partial_{\\theta \\theta \\theta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third poloidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 3, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _R_ttt(params, transforms, profiles, data, **kwargs):
    data["R_ttt"] = transforms["R"].transform(params["R_lmn"], 0, 3, 0)
    return data


@register_compute_fun(
    name="R_ttz",
    label="\\partial_{\\theta \\theta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative wrt poloidal angle "
    "twice and toroidal angle",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 2, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_ttz(params, transforms, profiles, data, **kwargs):
    data["R_ttz"] = transforms["R"].transform(params["R_lmn"], 0, 2, 1)
    return data


@register_compute_fun(
    name="R_tz",
    label="\\partial_{\\theta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second derivative wrt poloidal "
    "and toroidal angles",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_tz(params, transforms, profiles, data, **kwargs):
    data["R_tz"] = transforms["R"].transform(params["R_lmn"], 0, 1, 1)
    return data


@register_compute_fun(
    name="R_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third derivative wrt poloidal angle "
    "and toroidal angle twice",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 1, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_tzz(params, transforms, profiles, data, **kwargs):
    data["R_tzz"] = transforms["R"].transform(params["R_lmn"], 0, 1, 2)
    return data


@register_compute_fun(
    name="R_z",
    label="\\partial_{\\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, first toroidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_z(params, transforms, profiles, data, **kwargs):
    data["R_z"] = transforms["R"].transform(params["R_lmn"], 0, 0, 1)
    return data


@register_compute_fun(
    name="R_zz",
    label="\\partial_{\\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, second toroidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_zz(params, transforms, profiles, data, **kwargs):
    data["R_zz"] = transforms["R"].transform(params["R_lmn"], 0, 0, 2)
    return data


@register_compute_fun(
    name="R_zzz",
    label="\\partial_{\\zeta \\zeta \\zeta} R",
    units="m",
    units_long="meters",
    description="Major radius in lab frame, third toroidal derivative",
    dim=1,
    params=["R_lmn"],
    transforms={"R": [[0, 0, 3]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _R_zzz(params, transforms, profiles, data, **kwargs):
    data["R_zzz"] = transforms["R"].transform(params["R_lmn"], 0, 0, 3)
    return data


@register_compute_fun(
    name="X",
    label="X = R \\cos{\\phi}",
    units="m",
    units_long="meters",
    description="Cartesian X coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "phi"],
)
def _X(params, transforms, profiles, data, **kwargs):
    data["X"] = data["R"] * jnp.cos(data["phi"])
    return data


@register_compute_fun(
    name="X_r",
    label="\\partial_{\\rho} X",
    units="m",
    units_long="meters",
    description="Cartesian X coordinate, derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_r", "phi", "phi_r"],
)
def _X_r(params, transforms, profiles, data, **kwargs):
    data["X_r"] = (
        data["R_r"] * jnp.cos(data["phi"])
        - data["R"] * jnp.sin(data["phi"]) * data["phi_r"]
    )
    return data


@register_compute_fun(
    name="X_t",
    label="\\partial_{\\theta} X",
    units="m",
    units_long="meters",
    description="Cartesian X coordinate, derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_t", "phi", "phi_t"],
)
def _X_t(params, transforms, profiles, data, **kwargs):
    data["X_t"] = (
        data["R_t"] * jnp.cos(data["phi"])
        - data["R"] * jnp.sin(data["phi"]) * data["phi_t"]
    )
    return data


@register_compute_fun(
    name="X_z",
    label="\\partial_{\\zeta} X",
    units="m",
    units_long="meters",
    description="Cartesian X coordinate, derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_z", "phi", "phi_z"],
)
def _X_z(params, transforms, profiles, data, **kwargs):
    data["X_z"] = (
        data["R_z"] * jnp.cos(data["phi"])
        - data["R"] * jnp.sin(data["phi"]) * data["phi_z"]
    )
    return data


@register_compute_fun(
    name="Y",
    label="Y = R \\sin{\\phi}",
    units="m",
    units_long="meters",
    description="Cartesian Y coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "phi"],
)
def _Y(params, transforms, profiles, data, **kwargs):
    data["Y"] = data["R"] * jnp.sin(data["phi"])
    return data


@register_compute_fun(
    name="Y_r",
    label="\\partial_{\\rho} Y",
    units="m",
    units_long="meters",
    description="Cartesian Y coordinate, derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_r", "phi", "phi_r"],
)
def _Y_r(params, transforms, profiles, data, **kwargs):
    data["Y_r"] = (
        data["R_r"] * jnp.sin(data["phi"])
        + data["R"] * jnp.cos(data["phi"]) * data["phi_r"]
    )
    return data


@register_compute_fun(
    name="Y_t",
    label="\\partial_{\\theta} Y",
    units="m",
    units_long="meters",
    description="Cartesian Y coordinate, derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_t", "phi", "phi_t"],
)
def _Y_t(params, transforms, profiles, data, **kwargs):
    data["Y_t"] = (
        data["R_t"] * jnp.sin(data["phi"])
        + data["R"] * jnp.cos(data["phi"]) * data["phi_t"]
    )
    return data


@register_compute_fun(
    name="Y_z",
    label="\\partial_{\\zeta} Y",
    units="m",
    units_long="meters",
    description="Cartesian Y coordinate, derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["R", "R_z", "phi", "phi_z"],
)
def _Y_z(params, transforms, profiles, data, **kwargs):
    data["Y_z"] = (
        data["R_z"] * jnp.sin(data["phi"])
        + data["R"] * jnp.cos(data["phi"]) * data["phi_z"]
    )
    return data


@register_compute_fun(
    name="Z",
    label="Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z(params, transforms, profiles, data, **kwargs):
    data["Z"] = transforms["Z"].transform(params["Z_lmn"], 0, 0, 0)
    return data


@register_compute_fun(
    name="Z_r",
    label="\\partial_{\\rho} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, first radial derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_r(params, transforms, profiles, data, **kwargs):
    data["Z_r"] = transforms["Z"].transform(params["Z_lmn"], 1, 0, 0)
    return data


@register_compute_fun(
    name="Z_rr",
    label="\\partial_{\\rho \\rho} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second radial derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rr(params, transforms, profiles, data, **kwargs):
    data["Z_rr"] = transforms["Z"].transform(params["Z_lmn"], 2, 0, 0)
    return data


@register_compute_fun(
    name="Z_rrr",
    label="\\partial_{\\rho \\rho \\rho} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third radial derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[3, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rrr(params, transforms, profiles, data, **kwargs):
    data["Z_rrr"] = transforms["Z"].transform(params["Z_lmn"], 3, 0, 0)
    return data


@register_compute_fun(
    name="Z_rrrr",
    label="\\partial_{\\rho \\rho \\rho \\rho} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth radial derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[4, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rrrr(params, transforms, profiles, data, **kwargs):
    data["Z_rrrr"] = transforms["Z"].transform(params["Z_lmn"], 4, 0, 0)
    return data


@register_compute_fun(
    name="Z_rrrt",
    label="\\partial_{\\rho \\rho \\rho \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative wrt "
    " radial coordinate thrice and poloidal once",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[3, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rrrt(params, transforms, profiles, data, **kwargs):
    data["Z_rrrt"] = transforms["Z"].transform(params["Z_lmn"], 3, 1, 0)
    return data


@register_compute_fun(
    name="Z_rrrz",
    label="\\partial_{\\rho \\rho \\rho \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative wrt "
    " radial coordinate thrice and toroidal once",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[3, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rrrz(params, transforms, profiles, data, **kwargs):
    data["Z_rrrz"] = transforms["Z"].transform(params["Z_lmn"], 3, 0, 1)
    return data


@register_compute_fun(
    name="Z_rrt",
    label="\\partial_{\\rho \\rho \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative, wrt radius "
    "twice and poloidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rrt(params, transforms, profiles, data, **kwargs):
    data["Z_rrt"] = transforms["Z"].transform(params["Z_lmn"], 2, 1, 0)
    return data


@register_compute_fun(
    name="Z_rrtt",
    label="\\partial_{\\rho \\rho \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative, wrt radius "
    "twice and poloidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rrtt(params, transforms, profiles, data, **kwargs):
    data["Z_rrtt"] = transforms["Z"].transform(params["Z_lmn"], 2, 2, 0)
    return data


@register_compute_fun(
    name="Z_rrtz",
    label="\\partial_{\\rho \\rho \\theta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative wrt radius"
    "twice, poloidal angle, and toroidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rrtz(params, transforms, profiles, data, **kwargs):
    data["Z_rrtz"] = transforms["Z"].transform(params["Z_lmn"], 2, 1, 1)
    return data


@register_compute_fun(
    name="Z_rrz",
    label="\\partial_{\\rho \\rho \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative, wrt radius "
    "twice and toroidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rrz(params, transforms, profiles, data, **kwargs):
    data["Z_rrz"] = transforms["Z"].transform(params["Z_lmn"], 2, 0, 1)
    return data


@register_compute_fun(
    name="Z_rrzz",
    label="\\partial_{\\rho \\rho \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative, wrt radius "
    "twice and toroidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[2, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rrzz(params, transforms, profiles, data, **kwargs):
    data["Z_rrzz"] = transforms["Z"].transform(params["Z_lmn"], 2, 0, 2)
    return data


@register_compute_fun(
    name="Z_rt",
    label="\\partial_{\\rho \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second derivative wrt radius "
    "and poloidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rt(params, transforms, profiles, data, **kwargs):
    data["Z_rt"] = transforms["Z"].transform(params["Z_lmn"], 1, 1, 0)
    return data


@register_compute_fun(
    name="Z_rtt",
    label="\\partial_{\\rho \\theta \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt radius "
    "and poloidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rtt(params, transforms, profiles, data, **kwargs):
    data["Z_rtt"] = transforms["Z"].transform(params["Z_lmn"], 1, 2, 0)
    return data


@register_compute_fun(
    name="Z_rttt",
    label="\\partial_{\\rho \\theta \\theta \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt radius "
    "and poloidal angle thrice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 3, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
)
def _Z_rttt(params, transforms, profiles, data, **kwargs):
    data["Z_rttt"] = transforms["Z"].transform(params["Z_lmn"], 1, 3, 0)
    return data


@register_compute_fun(
    name="Z_rttz",
    label="\\partial_{\\rho \\theta \\theta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative wrt radius "
    "once, poloidal angle twice, and toroidal angle once",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 2, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rttz(params, transforms, profiles, data, **kwargs):
    data["Z_rttz"] = transforms["Z"].transform(params["Z_lmn"], 1, 2, 1)
    return data


@register_compute_fun(
    name="Z_rtz",
    label="\\partial_{\\rho \\theta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt radius, "
    "poloidal angle, and toroidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rtz(params, transforms, profiles, data, **kwargs):
    data["Z_rtz"] = transforms["Z"].transform(params["Z_lmn"], 1, 1, 1)
    return data


@register_compute_fun(
    name="Z_rtzz",
    label="\\partial_{\\rho \\theta \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, fourth derivative wrt radius, "
    "poloidal angle, and toroidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 1, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rtzz(params, transforms, profiles, data, **kwargs):
    data["Z_rtzz"] = transforms["Z"].transform(params["Z_lmn"], 1, 1, 2)
    return data


@register_compute_fun(
    name="Z_rz",
    label="\\partial_{\\rho \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second derivative wrt radius "
    "and toroidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rz(params, transforms, profiles, data, **kwargs):
    data["Z_rz"] = transforms["Z"].transform(params["Z_lmn"], 1, 0, 1)
    return data


@register_compute_fun(
    name="Z_rzz",
    label="\\partial_{\\rho \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt radius "
    "and toroidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rzz(params, transforms, profiles, data, **kwargs):
    data["Z_rzz"] = transforms["Z"].transform(params["Z_lmn"], 1, 0, 2)
    return data


@register_compute_fun(
    name="Z_rzzz",
    label="\\partial_{\\rho \\zeta \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt radius "
    "and toroidal angle thrice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[1, 0, 3]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _Z_rzzz(params, transforms, profiles, data, **kwargs):
    data["Z_rzzz"] = transforms["Z"].transform(params["Z_lmn"], 1, 0, 3)
    return data


@register_compute_fun(
    name="Z_t",
    label="\\partial_{\\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, first poloidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _Z_t(params, transforms, profiles, data, **kwargs):
    data["Z_t"] = transforms["Z"].transform(params["Z_lmn"], 0, 1, 0)
    return data


@register_compute_fun(
    name="Z_tt",
    label="\\partial_{\\theta \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second poloidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _Z_tt(params, transforms, profiles, data, **kwargs):
    data["Z_tt"] = transforms["Z"].transform(params["Z_lmn"], 0, 2, 0)
    return data


@register_compute_fun(
    name="Z_ttt",
    label="\\partial_{\\theta \\theta \\theta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third poloidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 3, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _Z_ttt(params, transforms, profiles, data, **kwargs):
    data["Z_ttt"] = transforms["Z"].transform(params["Z_lmn"], 0, 3, 0)
    return data


@register_compute_fun(
    name="Z_ttz",
    label="\\partial_{\\theta \\theta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt poloidal "
    "angle twice and toroidal angle",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 2, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_ttz(params, transforms, profiles, data, **kwargs):
    data["Z_ttz"] = transforms["Z"].transform(params["Z_lmn"], 0, 2, 1)
    return data


@register_compute_fun(
    name="Z_tz",
    label="\\partial_{\\theta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second derivative wrt poloidal "
    "and toroidal angles",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_tz(params, transforms, profiles, data, **kwargs):
    data["Z_tz"] = transforms["Z"].transform(params["Z_lmn"], 0, 1, 1)
    return data


@register_compute_fun(
    name="Z_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third derivative wrt poloidal "
    "angle and toroidal angle twice",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 1, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_tzz(params, transforms, profiles, data, **kwargs):
    data["Z_tzz"] = transforms["Z"].transform(params["Z_lmn"], 0, 1, 2)
    return data


@register_compute_fun(
    name="Z_z",
    label="\\partial_{\\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, first toroidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_z(params, transforms, profiles, data, **kwargs):
    data["Z_z"] = transforms["Z"].transform(params["Z_lmn"], 0, 0, 1)
    return data


@register_compute_fun(
    name="Z_zz",
    label="\\partial_{\\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, second toroidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_zz(params, transforms, profiles, data, **kwargs):
    data["Z_zz"] = transforms["Z"].transform(params["Z_lmn"], 0, 0, 2)
    return data


@register_compute_fun(
    name="Z_zzz",
    label="\\partial_{\\zeta \\zeta \\zeta} Z",
    units="m",
    units_long="meters",
    description="Vertical coordinate in lab frame, third toroidal derivative",
    dim=1,
    params=["Z_lmn"],
    transforms={"Z": [[0, 0, 3]]},
    profiles=[],
    coordinates="rtz",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.surface.FourierRZToroidalSurface",
    ],
)
def _Z_zzz(params, transforms, profiles, data, **kwargs):
    data["Z_zzz"] = transforms["Z"].transform(params["Z_lmn"], 0, 0, 3)
    return data


@register_compute_fun(
    name="alpha",
    label="\\alpha",
    units="~",
    units_long="None",
    description="Field line label",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST", "phi", "iota"],
)
def _alpha(params, transforms, profiles, data, **kwargs):
    data["alpha"] = data["theta_PEST"] - data["iota"] * data["phi"]
    return data


@register_compute_fun(
    name="alpha_r",
    label="\\partial_\\rho \\alpha",
    units="~",
    units_long="None",
    description="Field line label, derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_r", "phi", "phi_r", "iota", "iota_r"],
)
def _alpha_r(params, transforms, profiles, data, **kwargs):
    data["alpha_r"] = (
        data["theta_PEST_r"]
        - data["iota_r"] * data["phi"]
        - data["iota"] * data["phi_r"]
    )
    return data


@register_compute_fun(
    name="alpha_t",
    label="\\partial_\\theta \\alpha",
    units="~",
    units_long="None",
    description="Field line label, derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_t", "phi_t", "iota"],
)
def _alpha_t(params, transforms, profiles, data, **kwargs):
    data["alpha_t"] = data["theta_PEST_t"] - data["iota"] * data["phi_t"]
    return data


@register_compute_fun(
    name="alpha_tz",
    label="\\partial_{\\theta \\zeta} \\alpha",
    units="~",
    units_long="None",
    description="Field line label, derivative wrt poloidal and toroidal coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_tz", "phi_tz", "iota"],
)
def _alpha_tz(params, transforms, profiles, data, **kwargs):
    data["alpha_tz"] = data["theta_PEST_tz"] - data["iota"] * data["phi_tz"]
    return data


@register_compute_fun(
    name="alpha_z",
    label="\\partial_\\zeta \\alpha",
    units="~",
    units_long="None",
    description="Field line label, derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_z", "phi_z", "iota"],
)
def _alpha_z(params, transforms, profiles, data, **kwargs):
    data["alpha_z"] = data["theta_PEST_z"] - data["iota"] * data["phi_z"]
    return data


@register_compute_fun(
    name="alpha_tt",
    label="\\partial_{\\theta \\theta} \\alpha",
    units="~",
    units_long="None",
    description="Field line label, second-order derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["phi_tt", "iota", "theta_PEST_tt"],
)
def _alpha_tt(params, transforms, profiles, data, **kwargs):
    data["alpha_tt"] = data["theta_PEST_tt"] - data["iota"] * data["phi_tt"]
    return data


@register_compute_fun(
    name="alpha_zz",
    label="\\partial_{\\zeta \\zeta} \\alpha",
    units="~",
    units_long="None",
    description="Field line label, second-order derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta_PEST_zz", "phi_zz", "iota"],
)
def _alpha_zz(params, transforms, profiles, data, **kwargs):
    data["alpha_zz"] = data["theta_PEST_zz"] - data["iota"] * data["phi_zz"]
    return data


@register_compute_fun(
    name="lambda",
    label="\\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda(params, transforms, profiles, data, **kwargs):
    data["lambda"] = transforms["L"].transform(params["L_lmn"], 0, 0, 0)
    return data


@register_compute_fun(
    name="lambda_r",
    label="\\partial_{\\rho} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, first radial derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_r(params, transforms, profiles, data, **kwargs):
    data["lambda_r"] = transforms["L"].transform(params["L_lmn"], 1, 0, 0)
    return data


@register_compute_fun(
    name="lambda_rr",
    label="\\partial_{\\rho \\rho} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second radial derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[2, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rr(params, transforms, profiles, data, **kwargs):
    data["lambda_rr"] = transforms["L"].transform(params["L_lmn"], 2, 0, 0)
    return data


@register_compute_fun(
    name="lambda_rrr",
    label="\\partial_{\\rho \\rho \\rho} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third radial derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[3, 0, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rrr(params, transforms, profiles, data, **kwargs):
    data["lambda_rrr"] = transforms["L"].transform(params["L_lmn"], 3, 0, 0)
    return data


@register_compute_fun(
    name="lambda_rrrt",
    label="\\partial_{\\rho \\rho \\rho \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third radial derivative and"
    " first poloidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[3, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rrrt(params, transforms, profiles, data, **kwargs):
    data["lambda_rrrt"] = transforms["L"].transform(params["L_lmn"], 3, 1, 0)
    return data


@register_compute_fun(
    name="lambda_rrrz",
    label="\\partial_{\\rho \\rho \\rho \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third radial derivative and"
    " first toroidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[3, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rrrz(params, transforms, profiles, data, **kwargs):
    data["lambda_rrrz"] = transforms["L"].transform(params["L_lmn"], 3, 0, 1)
    return data


@register_compute_fun(
    name="lambda_rrt",
    label="\\partial_{\\rho \\rho \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative, wrt radius twice "
    "and poloidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[2, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rrt(params, transforms, profiles, data, **kwargs):
    data["lambda_rrt"] = transforms["L"].transform(params["L_lmn"], 2, 1, 0)
    return data


@register_compute_fun(
    name="lambda_rrz",
    label="\\partial_{\\rho \\rho \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative, wrt radius twice "
    "and toroidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[2, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rrz(params, transforms, profiles, data, **kwargs):
    data["lambda_rrz"] = transforms["L"].transform(params["L_lmn"], 2, 0, 1)
    return data


@register_compute_fun(
    name="lambda_rt",
    label="\\partial_{\\rho \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second derivative wrt radius and "
    "poloidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rt(params, transforms, profiles, data, **kwargs):
    data["lambda_rt"] = transforms["L"].transform(params["L_lmn"], 1, 1, 0)
    return data


@register_compute_fun(
    name="lambda_rtt",
    label="\\partial_{\\rho \\theta \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative wrt radius and "
    "poloidal angle twice",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rtt(params, transforms, profiles, data, **kwargs):
    data["lambda_rtt"] = transforms["L"].transform(params["L_lmn"], 1, 2, 0)
    return data


@register_compute_fun(
    name="lambda_rtz",
    label="\\partial_{\\rho \\theta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative wrt radius, poloidal "
    " angle, and toroidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rtz(params, transforms, profiles, data, **kwargs):
    data["lambda_rtz"] = transforms["L"].transform(params["L_lmn"], 1, 1, 1)
    return data


@register_compute_fun(
    name="lambda_rz",
    label="\\partial_{\\rho \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second derivative wrt radius and "
    "toroidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rz(params, transforms, profiles, data, **kwargs):
    data["lambda_rz"] = transforms["L"].transform(params["L_lmn"], 1, 0, 1)
    return data


@register_compute_fun(
    name="lambda_rzz",
    label="\\partial_{\\rho \\zeta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative wrt radius and "
    "toroidal angle twice",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[1, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_rzz(params, transforms, profiles, data, **kwargs):
    data["lambda_rzz"] = transforms["L"].transform(params["L_lmn"], 1, 0, 2)
    return data


@register_compute_fun(
    name="lambda_t",
    label="\\partial_{\\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, first poloidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 1, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_t(params, transforms, profiles, data, **kwargs):
    data["lambda_t"] = transforms["L"].transform(params["L_lmn"], 0, 1, 0)
    return data


@register_compute_fun(
    name="lambda_tt",
    label="\\partial_{\\theta \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second poloidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 2, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_tt(params, transforms, profiles, data, **kwargs):
    data["lambda_tt"] = transforms["L"].transform(params["L_lmn"], 0, 2, 0)
    return data


@register_compute_fun(
    name="lambda_ttt",
    label="\\partial_{\\theta \\theta \\theta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third poloidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 3, 0]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_ttt(params, transforms, profiles, data, **kwargs):
    data["lambda_ttt"] = transforms["L"].transform(params["L_lmn"], 0, 3, 0)
    return data


@register_compute_fun(
    name="lambda_ttz",
    label="\\partial_{\\theta \\theta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative wrt poloidal angle "
    "twice and toroidal angle",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 2, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_ttz(params, transforms, profiles, data, **kwargs):
    data["lambda_ttz"] = transforms["L"].transform(params["L_lmn"], 0, 2, 1)
    return data


@register_compute_fun(
    name="lambda_tz",
    label="\\partial_{\\theta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second derivative wrt poloidal and "
    "toroidal angles",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 1, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_tz(params, transforms, profiles, data, **kwargs):
    data["lambda_tz"] = transforms["L"].transform(params["L_lmn"], 0, 1, 1)
    return data


@register_compute_fun(
    name="lambda_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third derivative wrt poloidal angle "
    "and toroidal angle twice",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 1, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_tzz(params, transforms, profiles, data, **kwargs):
    data["lambda_tzz"] = transforms["L"].transform(params["L_lmn"], 0, 1, 2)
    return data


@register_compute_fun(
    name="lambda_z",
    label="\\partial_{\\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, first toroidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 0, 1]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_z(params, transforms, profiles, data, **kwargs):
    data["lambda_z"] = transforms["L"].transform(params["L_lmn"], 0, 0, 1)
    return data


@register_compute_fun(
    name="lambda_zz",
    label="\\partial_{\\zeta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, second toroidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 0, 2]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_zz(params, transforms, profiles, data, **kwargs):
    data["lambda_zz"] = transforms["L"].transform(params["L_lmn"], 0, 0, 2)
    return data


@register_compute_fun(
    name="lambda_zzz",
    label="\\partial_{\\zeta \\zeta \\zeta} \\lambda",
    units="rad",
    units_long="radians",
    description="Poloidal stream function, third toroidal derivative",
    dim=1,
    params=["L_lmn"],
    transforms={"L": [[0, 0, 3]]},
    profiles=[],
    coordinates="rtz",
    data=[],
)
def _lambda_zzz(params, transforms, profiles, data, **kwargs):
    data["lambda_zzz"] = transforms["L"].transform(params["L_lmn"], 0, 0, 3)
    return data


@register_compute_fun(
    name="omega",
    label="\\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function",
    dim=1,
    params=[],  # ["W_lmn"],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega(params, transforms, profiles, data, **kwargs):
    data["omega"] = data["0"]
    return data


@register_compute_fun(
    name="omega_r",
    label="\\partial_{\\rho} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, first radial derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_r(params, transforms, profiles, data, **kwargs):
    data["omega_r"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rr",
    label="\\partial_{\\rho \\rho} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second radial derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rr(params, transforms, profiles, data, **kwargs):
    data["omega_rr"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrr",
    label="\\partial_{\\rho \\rho \\rho} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third radial derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrr(params, transforms, profiles, data, **kwargs):
    data["omega_rrr"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrrr",
    label="\\partial_{\\rho \\rho \\rho \\rho} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth radial derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrrr(params, transforms, profiles, data, **kwargs):
    data["omega_rrrr"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrrt",
    label="\\partial_{\\rho \\rho \\rho \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative wrt radial coordinate"
    " thrice and poloidal once",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrrt(params, transforms, profiles, data, **kwargs):
    data["omega_rrrt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrrz",
    label="\\partial_{\\rho \\rho \\rho \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative wrt radial coordinate"
    " thrice and toroidal once",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrrz(params, transforms, profiles, data, **kwargs):
    data["omega_rrrz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrt",
    label="\\partial_{\\rho \\rho \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative, wrt radius twice "
    "and poloidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrt(params, transforms, profiles, data, **kwargs):
    data["omega_rrt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrtt",
    label="\\partial_{\\rho \\rho \\theta \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative, wrt radius twice "
    "and poloidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrtt(params, transforms, profiles, data, **kwargs):
    data["omega_rrtt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrtz",
    label="\\partial_{\\rho \\theta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative wrt radius twice,"
    " poloidal angle, and toroidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrtz(params, transforms, profiles, data, **kwargs):
    data["omega_rrtz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrz",
    label="\\partial_{\\rho \\rho \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative, wrt radius twice "
    "and toroidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrz(params, transforms, profiles, data, **kwargs):
    data["omega_rrz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rrzz",
    label="\\partial_{\\rho \\rho \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative, wrt radius twice "
    "and toroidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rrzz(params, transforms, profiles, data, **kwargs):
    data["omega_rrzz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rt",
    label="\\partial_{\\rho \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second derivative wrt radius and "
    "poloidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rt(params, transforms, profiles, data, **kwargs):
    data["omega_rt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rtt",
    label="\\partial_{\\rho \\theta \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt radius and "
    "poloidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rtt(params, transforms, profiles, data, **kwargs):
    data["omega_rtt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rttt",
    label="\\partial_{\\rho \\theta \\theta \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt radius and "
    "poloidal angle thrice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rttt(params, transforms, profiles, data, **kwargs):
    data["omega_rttt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rttz",
    label="\\partial_{\\rho \\theta \\theta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative wrt radius once, "
    "poloidal angle twice, and toroidal angle once",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rttz(params, transforms, profiles, data, **kwargs):
    data["omega_rttz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rtz",
    label="\\partial_{\\rho \\theta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt radius, poloidal"
    " angle, and toroidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rtz(params, transforms, profiles, data, **kwargs):
    data["omega_rtz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rtzz",
    label="\\partial_{\\rho \\theta \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, fourth derivative wrt radius, poloidal"
    " angle, and toroidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rtzz(params, transforms, profiles, data, **kwargs):
    data["omega_rtzz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rz",
    label="\\partial_{\\rho \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second derivative wrt radius and "
    "toroidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rz(params, transforms, profiles, data, **kwargs):
    data["omega_rz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rzz",
    label="\\partial_{\\rho \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt radius and "
    "toroidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rzz(params, transforms, profiles, data, **kwargs):
    data["omega_rzz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_rzzz",
    label="\\partial_{\\rho \\zeta \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt radius and "
    "toroidal angle thrice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_rzzz(params, transforms, profiles, data, **kwargs):
    data["omega_rzzz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_t",
    label="\\partial_{\\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, first poloidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_t(params, transforms, profiles, data, **kwargs):
    data["omega_t"] = data["0"]
    return data


@register_compute_fun(
    name="omega_tt",
    label="\\partial_{\\theta \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second poloidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_tt(params, transforms, profiles, data, **kwargs):
    data["omega_tt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_ttt",
    label="\\partial_{\\theta \\theta \\theta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third poloidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_ttt(params, transforms, profiles, data, **kwargs):
    data["omega_ttt"] = data["0"]
    return data


@register_compute_fun(
    name="omega_ttz",
    label="\\partial_{\\theta \\theta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt poloidal angle "
    "twice and toroidal angle",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_ttz(params, transforms, profiles, data, **kwargs):
    data["omega_ttz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_tz",
    label="\\partial_{\\theta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second derivative wrt poloidal and "
    "toroidal angles",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_tz(params, transforms, profiles, data, **kwargs):
    data["omega_tz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third derivative wrt poloidal angle "
    "and toroidal angle twice",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_tzz(params, transforms, profiles, data, **kwargs):
    data["omega_tzz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_z",
    label="\\partial_{\\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, first toroidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_z(params, transforms, profiles, data, **kwargs):
    data["omega_z"] = data["0"]
    return data


@register_compute_fun(
    name="omega_zz",
    label="\\partial_{\\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, second toroidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_zz(params, transforms, profiles, data, **kwargs):
    data["omega_zz"] = data["0"]
    return data


@register_compute_fun(
    name="omega_zzz",
    label="\\partial_{\\zeta \\zeta \\zeta} \\omega",
    units="rad",
    units_long="radians",
    description="Toroidal stream function, third toroidal derivative",
    dim=1,
    params=[],  # ["W_lmn"]
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["0"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _omega_zzz(params, transforms, profiles, data, **kwargs):
    data["omega_zzz"] = data["0"]
    return data


@register_compute_fun(
    name="phi",
    label="\\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["zeta", "omega"],
)
def _phi(params, transforms, profiles, data, **kwargs):
    data["phi"] = data["zeta"] + data["omega"]
    return data


@register_compute_fun(
    name="phi_r",
    label="\\partial_{\\rho} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_r"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _phi_r(params, transforms, profiles, data, **kwargs):
    data["phi_r"] = data["omega_r"]
    return data


@register_compute_fun(
    name="phi_rr",
    label="\\partial_{\\rho \\rho} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rr"],
)
def _phi_rr(params, transforms, profiles, data, **kwargs):
    data["phi_rr"] = data["omega_rr"]
    return data


@register_compute_fun(
    name="phi_rrz",
    label="\\partial_{\\rho \\rho \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt radial coordinate "
    "and first wrt DESC toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rrz"],
)
def _phi_rrz(params, transforms, profiles, data, **kwargs):
    data["phi_rrz"] = data["omega_rrz"]
    return data


@register_compute_fun(
    name="phi_rt",
    label="\\partial_{\\rho \\theta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt radial and "
    "poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rt"],
)
def _phi_rt(params, transforms, profiles, data, **kwargs):
    data["phi_rt"] = data["omega_rt"]
    return data


@register_compute_fun(
    name="phi_rtz",
    label="\\partial_{\\rho \\theta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, third derivative wrt radial, "
    "poloidal, and toroidal coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rtz"],
)
def _phi_rtz(params, transforms, profiles, data, **kwargs):
    data["phi_rtz"] = data["omega_rtz"]
    return data


@register_compute_fun(
    name="phi_rz",
    label="\\partial_{\\rho \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt radial and "
    "toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rz"],
)
def _phi_rz(params, transforms, profiles, data, **kwargs):
    data["phi_rz"] = data["omega_rz"]
    return data


@register_compute_fun(
    name="phi_rzz",
    label="\\partial_{\\rho \\zeta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, first derivative wrt radial and "
    "second derivative wrt DESC toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_rzz"],
)
def _phi_rzz(params, transforms, profiles, data, **kwargs):
    data["phi_rzz"] = data["omega_rzz"]
    return data


@register_compute_fun(
    name="phi_t",
    label="\\partial_{\\theta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_t"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _phi_t(params, transforms, profiles, data, **kwargs):
    data["phi_t"] = data["omega_t"]
    return data


@register_compute_fun(
    name="phi_tt",
    label="\\partial_{\\theta \\theta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt poloidal "
    "coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_tt"],
)
def _phi_tt(params, transforms, profiles, data, **kwargs):
    data["phi_tt"] = data["omega_tt"]
    return data


@register_compute_fun(
    name="phi_ttz",
    label="\\partial_{\\theta \\theta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt poloidal "
    "coordinate and first derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_ttz"],
)
def _phi_ttz(params, transforms, profiles, data, **kwargs):
    data["phi_ttz"] = data["omega_ttz"]
    return data


@register_compute_fun(
    name="phi_tz",
    label="\\partial_{\\theta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, derivative wrt poloidal and "
    "toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_tz"],
)
def _phi_tz(params, transforms, profiles, data, **kwargs):
    data["phi_tz"] = data["omega_tz"]
    return data


@register_compute_fun(
    name="phi_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, derivative wrt poloidal coordinate and "
    "second derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_tzz"],
)
def _phi_tzz(params, transforms, profiles, data, **kwargs):
    data["phi_tzz"] = data["omega_tzz"]
    return data


@register_compute_fun(
    name="phi_z",
    label="\\partial_{\\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_z"],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _phi_z(params, transforms, profiles, data, **kwargs):
    data["phi_z"] = 1 + data["omega_z"]
    return data


@register_compute_fun(
    name="phi_zz",
    label="\\partial_{\\zeta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, second derivative wrt toroidal "
    "coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_zz"],
)
def _phi_zz(params, transforms, profiles, data, **kwargs):
    data["phi_zz"] = data["omega_zz"]
    return data


@register_compute_fun(
    name="phi_zzz",
    label="\\partial_{\\zeta \\zeta \\zeta} \\phi",
    units="rad",
    units_long="radians",
    description="Toroidal angle in lab frame, third derivative wrt toroidal "
    "coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["omega_zzz"],
)
def _phi_zzz(params, transforms, profiles, data, **kwargs):
    data["phi_zzz"] = data["omega_zzz"]
    return data


@register_compute_fun(
    name="rho",
    label="\\rho",
    units="~",
    units_long="None",
    description="Radial coordinate, proportional to the square root "
    + "of the toroidal flux",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
        "desc.magnetic_fields._core.OmnigenousField",
    ],
)
def _rho(params, transforms, profiles, data, **kwargs):
    data["rho"] = transforms["grid"].nodes[:, 0]
    return data


@register_compute_fun(
    name="theta",
    label="\\theta",
    units="rad",
    units_long="radians",
    description="Poloidal angular coordinate (geometric, not magnetic)",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="t",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _theta(params, transforms, profiles, data, **kwargs):
    data["theta"] = transforms["grid"].nodes[:, 1]
    return data


@register_compute_fun(
    name="theta_PEST",
    label="\\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["theta", "lambda"],
)
def _theta_PEST(params, transforms, profiles, data, **kwargs):
    data["theta_PEST"] = data["theta"] + data["lambda"]
    return data


@register_compute_fun(
    name="theta_PEST_r",
    label="\\partial_{\\rho} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_r"],
)
def _theta_PEST_r(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_r"] = data["lambda_r"]
    return data


@register_compute_fun(
    name="theta_PEST_rt",
    label="\\partial_{\\rho \\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate,"
    "derivative wrt poloidal and radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_rt"],
)
def _theta_PEST_rt(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_rt"] = data["lambda_rt"]
    return data


@register_compute_fun(
    name="theta_PEST_rrt",
    label="\\partial_{\\rho \\rho \\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, second "
    "derivative wrt radial coordinate and first derivative wrt DESC poloidal "
    "coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_rrt"],
)
def _theta_PEST_rrt(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_rrt"] = data["lambda_rrt"]
    return data


@register_compute_fun(
    name="theta_PEST_rtz",
    label="\\partial_{\\rho \\theta \\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "radial and DESC poloidal and toroidal coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_rtz"],
)
def _theta_PEST_rtz(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_rtz"] = data["lambda_rtz"]
    return data


@register_compute_fun(
    name="theta_PEST_rtt",
    label="\\partial_{\\rho \\theta \\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "radial coordinate once and DESC poloidal coordinate twice",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_rtt"],
)
def _theta_PEST_rtt(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_rtt"] = data["lambda_rtt"]
    return data


@register_compute_fun(
    name="theta_PEST_t",
    label="\\partial_{\\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_t"],
)
def _theta_PEST_t(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_t"] = 1 + data["lambda_t"]
    return data


@register_compute_fun(
    name="theta_PEST_tt",
    label="\\partial_{\\theta \\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate,"
    "second derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_tt"],
)
def _theta_PEST_tt(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_tt"] = data["lambda_tt"]
    return data


@register_compute_fun(
    name="theta_PEST_ttt",
    label="\\partial_{\\theta \\theta \\theta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, third "
    "derivative wrt poloidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_ttt"],
)
def _theta_PEST_ttt(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_ttt"] = data["lambda_ttt"]
    return data


@register_compute_fun(
    name="theta_PEST_tz",
    label="\\partial_{\\theta \\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "poloidal and toroidal coordinates",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_tz"],
)
def _theta_PEST_tz(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_tz"] = data["lambda_tz"]
    return data


@register_compute_fun(
    name="theta_PEST_tzz",
    label="\\partial_{\\theta \\zeta \\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, derivative wrt "
    "poloidal coordinate once and toroidal coordinate twice",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_tzz"],
)
def _theta_PEST_tzz(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_tzz"] = data["lambda_tzz"]
    return data


@register_compute_fun(
    name="theta_PEST_z",
    label="\\partial_{\\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate,"
    " derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_z"],
)
def _theta_PEST_z(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_z"] = data["lambda_z"]
    return data


@register_compute_fun(
    name="theta_PEST_zz",
    label="\\partial_{\\zeta \\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, second "
    "derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_zz"],
)
def _theta_PEST_zz(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_zz"] = data["lambda_zz"]
    return data


@register_compute_fun(
    name="theta_PEST_ttz",
    label="\\partial_{\\theta \\theta \\zeta} \\vartheta",
    units="rad",
    units_long="radians",
    description="PEST straight field line poloidal angular coordinate, second "
    "derivative wrt poloidal coordinate and derivative wrt toroidal coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["lambda_ttz"],
)
def _theta_PEST_ttz(params, transforms, profiles, data, **kwargs):
    data["theta_PEST_ttz"] = data["lambda_ttz"]
    return data


@register_compute_fun(
    name="zeta",
    label="\\zeta",
    units="rad",
    units_long="radians",
    description="Toroidal angular coordinate",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="z",
    data=[],
    parameterization=[
        "desc.equilibrium.equilibrium.Equilibrium",
        "desc.geometry.core.Surface",
    ],
)
def _zeta(params, transforms, profiles, data, **kwargs):
    data["zeta"] = transforms["grid"].nodes[:, 2]
    return data



from desc.backend import jnp

from .data_index import register_compute_fun
from .geom_utils import rpz2xyz

# TODO(#568): review when zeta no longer equals phi


@register_compute_fun(
    name="x",
    label="\\mathbf{r}",
    units="m",
    units_long="meters",
    description="Position vector along surface",
    dim=3,
    params=["R_lmn", "Z_lmn"],
    transforms={
        "R": [[0, 0, 0]],
        "Z": [[0, 0, 0]],
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
)
def _x_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    R = transforms["R"].transform(params["R_lmn"])
    Z = transforms["Z"].transform(params["Z_lmn"])
    # TODO(#568): change when zeta no longer equals phi
    phi = transforms["grid"].nodes[:, 2]
    coords = jnp.stack([R, phi, Z], axis=1)
    # default basis for "x" is rpz, the conversion will be done
    # in the main _compute function
    data["x"] = coords
    return data


@register_compute_fun(
    name="X",
    label="X",
    units="m",
    units_long="meters",
    description="Cartesian X coordinate.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["x"],
    parameterization="desc.geometry.core.Surface",
)
def _X_Surface(params, transforms, profiles, data, **kwargs):
    coords = data["x"]
    coords = rpz2xyz(coords)
    data["X"] = coords[:, 0]
    return data


@register_compute_fun(
    name="Y",
    label="Y",
    units="m",
    units_long="meters",
    description="Cartesian Y coordinate.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["x"],
    parameterization="desc.geometry.core.Surface",
)
def _Y_Surface(params, transforms, profiles, data, **kwargs):
    coords = data["x"]
    coords = rpz2xyz(coords)
    data["Y"] = coords[:, 1]
    return data


@register_compute_fun(
    name="R",
    label="R",
    units="m",
    units_long="meters",
    description="Cylindrical radial position along surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["x"],
    parameterization="desc.geometry.core.Surface",
)
def _R_Surface(params, transforms, profiles, data, **kwargs):
    coords = data["x"]
    data["R"] = coords[:, 0]
    return data


@register_compute_fun(
    name="phi",
    label="\\phi",
    units="rad",
    units_long="radians",
    description="Toroidal phi position along surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["x"],
    parameterization="desc.geometry.core.Surface",
)
def _phi_Surface(params, transforms, profiles, data, **kwargs):
    coords = data["x"]
    data["phi"] = coords[:, 1]
    return data


@register_compute_fun(
    name="Z",
    label="Z",
    units="m",
    units_long="meters",
    description="Cylindrical vertical position along surface",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["x"],
    parameterization="desc.geometry.core.Surface",
)
def _Z_Surface(params, transforms, profiles, data, **kwargs):
    data["Z"] = data["x"][:, 2]
    return data


@register_compute_fun(
    name="e_rho",
    label="\\mathbf{e}_{\\rho}",
    units="m",
    units_long="meters",
    description="Covariant radial basis vector",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
)
def _e_rho_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_rho"] = coords
    return data


@register_compute_fun(
    name="e_rho_r",
    label="\\partial_{\\rho} \\mathbf{e}_{\\rho}",
    units="m",
    units_long="meters",
    description="Covariant radial basis vector, derivative wrt radial coordinate",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
)
def _e_rho_r_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_rho_r"] = coords
    return data


@register_compute_fun(
    name="e_rho_rr",
    label="\\partial_{\\rho \\rho} \\mathbf{e}_{\\rho}",
    units="m",
    units_long="meters",
    description="Covariant radial basis vector,"
    " second derivative wrt radial coordinate",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
)
def _e_rho_rr_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_rho_rr"] = coords
    return data


@register_compute_fun(
    name="e_rho_t",
    label="\\partial_{\\theta} \\mathbf{e}_{\\rho}",
    units="m",
    units_long="meters",
    description="Covariant radial basis vector, derivative wrt poloidal angle",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
    aliases=["e_theta_r"],
)
def _e_rho_t_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_rho_t"] = coords
    return data


@register_compute_fun(
    name="e_rho_z",
    label="\\partial_{\\zeta} \\mathbf{e}_{\\rho}",
    units="m",
    units_long="meters",
    description="Covariant radial basis vector, derivative wrt toroidal angle",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization=[
        "desc.geometry.surface.FourierRZToroidalSurface",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
    aliases=["e_zeta_r"],
)
def _e_rho_z_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_rho_z"] = coords
    return data


@register_compute_fun(
    name="e_theta_rr",
    label="\\partial_{\\rho \\rho} \\mathbf{e}_{\\theta}",
    units="m",
    units_long="meters",
    description="Covariant poloidal basis vector,"
    " second derivative wrt radial coordinate",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.geometry.surface.FourierRZToroidalSurface",
    aliases=["e_rho_rt"],
)
def _e_theta_rr_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_theta_rr"] = coords
    return data


@register_compute_fun(
    name="e_zeta_rr",
    label="\\partial_{\\rho \\rho} \\mathbf{e}_{\\zeta}",
    units="m",
    units_long="meters",
    description="Covariant toroidal basis vector,"
    " second derivative wrt radial coordinate",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization=[
        "desc.geometry.surface.FourierRZToroidalSurface",
        "desc.geometry.surface.ZernikeRZToroidalSection",
    ],
    aliases=["e_rho_rz"],
)
def _e_zeta_rr_FourierRZToroidalSurface(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_zeta_rr"] = coords
    return data


@register_compute_fun(
    name="Phi",
    label="\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential",
    dim=1,
    params=["I", "G", "Phi_mn"],
    transforms={"Phi": [[0, 0, 0]]},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization=[
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField"
    ],
)
def _Phi_FourierCurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi"] = (
        transforms["Phi"].transform(params["Phi_mn"])
        + params["G"] * transforms["Phi"].nodes[:, 2].flatten(order="F") / 2 / jnp.pi
        + params["I"] * transforms["Phi"].nodes[:, 1].flatten(order="F") / 2 / jnp.pi
    )
    return data


@register_compute_fun(
    name="Phi_t",
    label="\\partial_{\\theta}\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential, poloidal derivative",
    dim=1,
    params=["I", "Phi_mn"],
    transforms={"Phi": [[0, 1, 0]]},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization=[
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField"
    ],
)
def _Phi_t_FourierCurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi_t"] = (
        transforms["Phi"].transform(params["Phi_mn"], dt=1) + params["I"] / 2 / jnp.pi
    )
    return data


@register_compute_fun(
    name="Phi_z",
    label="\\partial_{\\zeta}\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential, toroidal derivative",
    dim=1,
    params=["G", "Phi_mn"],
    transforms={"Phi": [[0, 0, 1]]},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization=[
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField"
    ],
)
def _Phi_z_FourierCurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi_z"] = (
        transforms["Phi"].transform(params["Phi_mn"], dz=1) + params["G"] / 2 / jnp.pi
    )
    return data


@register_compute_fun(
    name="Phi",
    label="\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential",
    dim=1,
    params=[],
    transforms={"grid": [], "potential": [], "params": []},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.magnetic_fields._current_potential.CurrentPotentialField",
)
def _Phi_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi"] = transforms["potential"](
        transforms["grid"].nodes[:, 1],
        transforms["grid"].nodes[:, 2],
        **transforms["params"]
    )
    return data


@register_compute_fun(
    name="Phi_t",
    label="\\partial_{\\theta}\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential, poloidal derivative",
    dim=1,
    params=[],
    transforms={"grid": [], "potential_dtheta": [], "params": []},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.magnetic_fields._current_potential.CurrentPotentialField",
)
def _Phi_t_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi_t"] = transforms["potential_dtheta"](
        transforms["grid"].nodes[:, 1],
        transforms["grid"].nodes[:, 2],
        **transforms["params"]
    )
    return data


@register_compute_fun(
    name="Phi_z",
    label="\\partial_{\\zeta}\\Phi",
    units="A",
    units_long="Amperes",
    description="Surface current potential, toroidal derivative",
    dim=1,
    params=[],
    transforms={"grid": [], "potential_dzeta": [], "params": []},
    profiles=[],
    coordinates="tz",
    data=[],
    parameterization="desc.magnetic_fields._current_potential.CurrentPotentialField",
)
def _Phi_z_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["Phi_z"] = transforms["potential_dzeta"](
        transforms["grid"].nodes[:, 1],
        transforms["grid"].nodes[:, 2],
        **transforms["params"]
    )
    return data


@register_compute_fun(
    name="K^theta",
    label="K^{\\theta}",
    units="A/m^2",
    units_long="Amperes per square meter",
    description="Contravariant poloidal component of surface current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="tz",
    data=["Phi_z", "|e_theta x e_zeta|"],
    parameterization=[
        "desc.magnetic_fields._current_potential.CurrentPotentialField",
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField",
    ],
)
def _K_sup_theta_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["K^theta"] = -data["Phi_z"] * (1 / data["|e_theta x e_zeta|"])
    return data


@register_compute_fun(
    name="K^zeta",
    label="K^{\\zeta}",
    units="A/m^2",
    units_long="Amperes per square meter",
    description="Contravariant toroidal component of surface current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="tz",
    data=["Phi_t", "|e_theta x e_zeta|"],
    parameterization=[
        "desc.magnetic_fields._current_potential.CurrentPotentialField",
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField",
    ],
)
def _K_sup_zeta_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["K^zeta"] = data["Phi_t"] * (1 / data["|e_theta x e_zeta|"])
    return data


@register_compute_fun(
    name="K",
    label="\\mathbf{K}",
    units="A/m",
    units_long="Amperes per meter",
    description="Surface current density, defined as the"
    "surface normal vector cross the gradient of the current potential.",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="tz",
    data=["K^theta", "K^zeta", "e_theta", "e_zeta"],
    parameterization=[
        "desc.magnetic_fields._current_potential.CurrentPotentialField",
        "desc.magnetic_fields._current_potential.FourierCurrentPotentialField",
    ],
)
def _K_CurrentPotentialField(params, transforms, profiles, data, **kwargs):
    data["K"] = (data["K^zeta"] * data["e_zeta"].T).T + (
        data["K^theta"] * data["e_theta"].T
    ).T
    return data


@register_compute_fun(
    name="x",
    label="\\mathbf{r}",
    units="m",
    units_long="meters",
    description="Position vector along surface",
    dim=3,
    params=["R_lmn", "Z_lmn"],
    transforms={
        "R": [[0, 0, 0]],
        "Z": [[0, 0, 0]],
        "grid": [],
    },
    profiles=[],
    coordinates="rt",
    data=[],
    parameterization="desc.geometry.surface.ZernikeRZToroidalSection",
)
def _x_ZernikeRZToroidalSection(params, transforms, profiles, data, **kwargs):
    R = transforms["R"].transform(params["R_lmn"])
    Z = transforms["Z"].transform(params["Z_lmn"])
    phi = transforms["grid"].nodes[:, 2]
    coords = jnp.stack([R, phi, Z], axis=1)
    data["x"] = coords
    return data


@register_compute_fun(
    name="e_zeta",
    label="\\mathbf{e}_{\\zeta}",
    units="m",
    units_long="meters",
    description="Covariant toroidal basis vector",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="rt",
    data=[],
    parameterization="desc.geometry.surface.ZernikeRZToroidalSection",
)
def _e_zeta_ZernikeRZToroidalSection(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_zeta"] = coords
    return data


@register_compute_fun(
    name="e_theta_z",
    label="\\partial_{\\zeta} \\mathbf{e}_{\\theta}",
    units="m",
    units_long="meters",
    description="Covariant poloidal basis vector, derivative wrt toroidal angle",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="rt",
    data=[],
    parameterization="desc.geometry.surface.ZernikeRZToroidalSection",
    aliases=["e_zeta_t"],
)
def _e_theta_z_ZernikeRZToroidalSection(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_theta_z"] = coords
    return data


@register_compute_fun(
    name="e_zeta_z",
    label="\\partial_{\\zeta} \\mathbf{e}_{\\zeta}",
    units="m",
    units_long="meters",
    description="Covariant toroidal basis vector, derivative wrt toroidal angle",
    dim=3,
    params=[],
    transforms={
        "grid": [],
    },
    profiles=[],
    coordinates="rt",
    data=[],
    parameterization="desc.geometry.surface.ZernikeRZToroidalSection",
)
def _e_zeta_z_ZernikeRZToroidalSection(params, transforms, profiles, data, **kwargs):
    coords = jnp.zeros((transforms["grid"].num_nodes, 3))
    data["e_zeta_z"] = coords
    return data



"""Compute functions for equilibrium objectives, i.e. Force and MHD energy.

Notes
-----
Some quantities require additional work to compute at the magnetic axis.
A Python lambda function is used to lazily compute the magnetic axis limits
of these quantities. These lambda functions are evaluated only when the
computational grid has a node on the magnetic axis to avoid potentially
expensive computations.
"""

from interpax import interp1d
from scipy.constants import elementary_charge, mu_0

from desc.backend import jnp

from ..integrals.surface_integral import surface_averages
from ..utils import cross, dot, safediv, safenorm
from .data_index import register_compute_fun


@register_compute_fun(
    name="J^rho",
    label="J^{\\rho}",
    units="A \\cdot m^{-3}",
    units_long="Amperes / cubic meter",
    description="Contravariant radial component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)", "B_zeta_t", "B_theta_z"],
    axis_limit_data=["sqrt(g)_r", "B_zeta_rt", "B_theta_rz"],
    parameterization="desc.equilibrium.equilibrium.Equilibrium",
)
def _J_sup_rho(params, transforms, profiles, data, **kwargs):
    # At the magnetic axis,
    # ∂_θ (𝐁 ⋅ 𝐞_ζ) - ∂_ζ (𝐁 ⋅ 𝐞_θ) = 𝐁 ⋅ (∂_θ 𝐞_ζ - ∂_ζ 𝐞_θ) = 0
    # because the partial derivatives commute. So 𝐉^ρ is of the indeterminate
    # form 0/0 and we may compute the limit as follows.
    data["J^rho"] = (
        transforms["grid"].replace_at_axis(
            safediv(data["B_zeta_t"] - data["B_theta_z"], data["sqrt(g)"]),
            lambda: safediv(data["B_zeta_rt"] - data["B_theta_rz"], data["sqrt(g)_r"]),
        )
    ) / mu_0
    return data


@register_compute_fun(
    name="J^theta*sqrt(g)",
    label="J^{\\theta} \\sqrt{g}",
    units="A",
    units_long="Amperes",
    description="Contravariant poloidal component of plasma current density,"
    " weighted by 3-D volume Jacobian",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B_rho_z", "B_zeta_r"],
)
def _J_sup_theta_sqrt_g(params, transforms, profiles, data, **kwargs):
    data["J^theta*sqrt(g)"] = (data["B_rho_z"] - data["B_zeta_r"]) / mu_0
    return data


@register_compute_fun(
    name="J^theta",
    label="J^{\\theta}",
    units="A \\cdot m^{-3}",
    units_long="Amperes / cubic meter",
    description="Contravariant poloidal component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)", "J^theta*sqrt(g)"],
)
def _J_sup_theta(params, transforms, profiles, data, **kwargs):
    data["J^theta"] = data["J^theta*sqrt(g)"] / data["sqrt(g)"]
    return data


@register_compute_fun(
    name="J^zeta",
    label="J^{\\zeta}",
    units="A \\cdot m^{-3}",
    units_long="Amperes / cubic meter",
    description="Contravariant toroidal component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["sqrt(g)", "B_theta_r", "B_rho_t"],
    axis_limit_data=["sqrt(g)_r", "B_theta_rr", "B_rho_rt"],
)
def _J_sup_zeta(params, transforms, profiles, data, **kwargs):
    # At the magnetic axis,
    # ∂ᵨ (𝐁 ⋅ 𝐞_θ) - ∂_θ (𝐁 ⋅ 𝐞ᵨ) = 𝐁 ⋅ (∂ᵨ 𝐞_θ - ∂_θ 𝐞ᵨ) = 0
    # because the partial derivatives commute. So 𝐉^ζ is of the indeterminate
    # form 0/0 and we may compute the limit as follows.
    data["J^zeta"] = (
        transforms["grid"].replace_at_axis(
            safediv(data["B_theta_r"] - data["B_rho_t"], data["sqrt(g)"]),
            lambda: safediv(data["B_theta_rr"] - data["B_rho_rt"], data["sqrt(g)_r"]),
        )
    ) / mu_0
    return data


@register_compute_fun(
    name="J",
    label="\\mathbf{J}",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Plasma current density",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "J^rho",
        "J^zeta",
        "J^theta*sqrt(g)",
        "e_rho",
        "e_zeta",
        "e_theta/sqrt(g)",
    ],
)
def _J(params, transforms, profiles, data, **kwargs):
    data["J"] = (
        data["J^rho"] * data["e_rho"].T
        + data["J^theta*sqrt(g)"] * data["e_theta/sqrt(g)"].T
        + data["J^zeta"] * data["e_zeta"].T
    ).T
    return data


@register_compute_fun(
    name="J*sqrt(g)",
    label="\\mathbf{J} \\sqrt{g}",
    units="A m",
    units_long="Ampere meters",
    description="Plasma current density weighted by 3-D volume Jacobian",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "B_rho_z",
        "B_theta_r",
        "B_zeta_t",
        "B_rho_t",
        "B_theta_z",
        "B_zeta_r",
        "e_rho",
        "e_theta",
        "e_zeta",
    ],
)
def _J_sqrt_g(params, transforms, profiles, data, **kwargs):
    data["J*sqrt(g)"] = (
        (data["B_zeta_t"] - data["B_theta_z"]) * data["e_rho"].T
        + (data["B_rho_z"] - data["B_zeta_r"]) * data["e_theta"].T
        + (data["B_theta_r"] - data["B_rho_t"]) * data["e_zeta"].T
    ).T / mu_0
    return data


@register_compute_fun(
    name="(J*sqrt(g))_r",
    label="\\partial_{\\rho} (\\mathbf{J} \\sqrt{g})",
    units="A m",
    units_long="Ampere meters",
    description="Plasma current density weighted by 3-D volume Jacobian,"
    " radial derivative",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=[
        "B_rho_z",
        "B_rho_rz",
        "B_theta_r",
        "B_theta_rr",
        "B_zeta_t",
        "B_zeta_rt",
        "B_rho_t",
        "B_rho_rt",
        "B_theta_z",
        "B_theta_rz",
        "B_zeta_r",
        "B_zeta_rr",
        "e_rho",
        "e_theta",
        "e_zeta",
        "e_rho_r",
        "e_theta_r",
        "e_zeta_r",
    ],
)
def _J_sqrt_g_r(params, transforms, profiles, data, **kwargs):
    data["(J*sqrt(g))_r"] = (
        (data["B_zeta_rt"] - data["B_theta_rz"]) * data["e_rho"].T
        + (data["B_zeta_t"] - data["B_theta_z"]) * data["e_rho_r"].T
        + (data["B_rho_rz"] - data["B_zeta_rr"]) * data["e_theta"].T
        + (data["B_rho_z"] - data["B_zeta_r"]) * data["e_theta_r"].T
        + (data["B_theta_rr"] - data["B_rho_rt"]) * data["e_zeta"].T
        + (data["B_theta_r"] - data["B_rho_t"]) * data["e_zeta_r"].T
    ).T / mu_0
    return data


@register_compute_fun(
    name="J_R",
    label="J_{R}",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Radial component of plasma current density in lab frame",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J"],
)
def _J_R(params, transforms, profiles, data, **kwargs):
    data["J_R"] = data["J"][:, 0]
    return data


@register_compute_fun(
    name="J_phi",
    label="J_{\\phi}",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Toroidal component of plasma current density in lab frame",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J"],
)
def _J_phi(params, transforms, profiles, data, **kwargs):
    data["J_phi"] = data["J"][:, 1]
    return data


@register_compute_fun(
    name="J_Z",
    label="J_{Z}",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Vertical component of plasma current density in lab frame",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J"],
)
def _J_Z(params, transforms, profiles, data, **kwargs):
    data["J_Z"] = data["J"][:, 2]
    return data


@register_compute_fun(
    name="|J|",
    label="|\\mathbf{J}|",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Magnitude of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J"],
)
def _Jmag(params, transforms, profiles, data, **kwargs):
    data["|J|"] = safenorm(data["J"], axis=-1)
    return data


@register_compute_fun(
    name="J_rho",
    label="J_{\\rho}",
    units="A / m",
    units_long="Amperes / meter",
    description="Covariant radial component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J", "e_rho"],
)
def _J_sub_rho(params, transforms, profiles, data, **kwargs):
    data["J_rho"] = dot(data["J"], data["e_rho"])
    return data


@register_compute_fun(
    name="J_theta",
    label="J_{\\theta}",
    units="A / m",
    units_long="Amperes / meter",
    description="Covariant poloidal component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J", "e_theta"],
)
def _J_sub_theta(params, transforms, profiles, data, **kwargs):
    data["J_theta"] = dot(data["J"], data["e_theta"])
    return data


@register_compute_fun(
    name="J_zeta",
    label="J_{\\zeta}",
    units="A / m",
    units_long="Amperes / meter",
    description="Covariant toroidal component of plasma current density",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J", "e_zeta"],
)
def _J_sub_zeta(params, transforms, profiles, data, **kwargs):
    data["J_zeta"] = dot(data["J"], data["e_zeta"])
    return data


@register_compute_fun(
    name="J*B",
    label="\\mathbf{J} \\cdot \\mathbf{B}",
    units="N / m^{3}",
    units_long="Newtons / cubic meter",
    description="Current density parallel to magnetic field, times field strength "
    + "(note units are not Amperes)",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J", "B"],
)
def _J_dot_B(params, transforms, profiles, data, **kwargs):
    data["J*B"] = dot(data["J"], data["B"])
    return data


@register_compute_fun(
    name="<J*B>",
    label="\\langle \\mathbf{J} \\cdot \\mathbf{B} \\rangle",
    units="N / m^{3}",
    units_long="Newtons / cubic meter",
    description="Flux surface average of current density dotted into magnetic field "
    + "(note units are not Amperes)",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["J*sqrt(g)", "B", "V_r(r)"],
    axis_limit_data=["(J*sqrt(g))_r", "V_rr(r)"],
    resolution_requirement="tz",
)
def _J_dot_B_fsa(params, transforms, profiles, data, **kwargs):
    J = transforms["grid"].replace_at_axis(
        data["J*sqrt(g)"], lambda: data["(J*sqrt(g))_r"], copy=True
    )
    data["<J*B>"] = surface_averages(
        transforms["grid"],
        dot(J, data["B"]),  # sqrt(g) factor pushed into J
        denominator=transforms["grid"].replace_at_axis(
            data["V_r(r)"], lambda: data["V_rr(r)"], copy=True
        ),
    )
    return data


@register_compute_fun(
    name="J_parallel",
    label="\\mathbf{J} \\cdot \\hat{\\mathbf{b}}",
    units="A \\cdot m^{-2}",
    units_long="Amperes / square meter",
    description="Plasma current density parallel to magnetic field",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J*B", "|B|"],
)
def _J_parallel(params, transforms, profiles, data, **kwargs):
    data["J_parallel"] = data["J*B"] / data["|B|"]
    return data


@register_compute_fun(
    name="F_rho",
    label="F_{\\rho}",
    units="N \\cdot m^{-2}",
    units_long="Newtons / square meter",
    description="Covariant radial component of force balance error",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["p_r", "(curl(B)xB)_rho"],
)
def _F_rho(params, transforms, profiles, data, **kwargs):
    data["F_rho"] = data["(curl(B)xB)_rho"] / mu_0 - data["p_r"]
    return data


@register_compute_fun(
    name="F_theta",
    label="F_{\\theta}",
    units="N \\cdot m^{-2}",
    units_long="Newtons / square meter",
    description="Covariant poloidal component of force balance error",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["F_helical", "B^zeta"],
)
def _F_theta(params, transforms, profiles, data, **kwargs):
    data["F_theta"] = -data["B^zeta"] * data["F_helical"]
    return data


@register_compute_fun(
    name="F_zeta",
    label="F_{\\zeta}",
    units="N \\cdot m^{-2}",
    units_long="Newtons / square meter",
    description="Covariant toroidal component of force balance error",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B^theta", "F_helical"],
)
def _F_zeta(params, transforms, profiles, data, **kwargs):
    data["F_zeta"] = data["B^theta"] * data["F_helical"]
    return data


@register_compute_fun(
    name="F_helical",
    label="F_{\\mathrm{helical}}",
    units="A",
    units_long="Amperes",
    description="Covariant helical component of force balance error",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B_zeta_t", "B_theta_z"],
)
def _F_helical(params, transforms, profiles, data, **kwargs):
    data["F_helical"] = (data["B_zeta_t"] - data["B_theta_z"]) / mu_0
    return data


@register_compute_fun(
    name="F",
    label="\\mathbf{J} \\times \\mathbf{B} - \\nabla p",
    units="N \\cdot m^{-3}",
    units_long="Newtons / cubic meter",
    description="Force balance error",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["F_rho", "F_zeta", "e^rho", "e^zeta", "B^zeta", "J^rho", "e^theta*sqrt(g)"],
)
def _F(params, transforms, profiles, data, **kwargs):
    # F_theta e^theta refactored as below to resolve indeterminacy at axis.
    data["F"] = (
        data["F_rho"] * data["e^rho"].T
        - data["B^zeta"] * data["J^rho"] * data["e^theta*sqrt(g)"].T
        + data["F_zeta"] * data["e^zeta"].T
    ).T
    return data


@register_compute_fun(
    name="|F|",
    label="|\\mathbf{J} \\times \\mathbf{B} - \\nabla p|",
    units="N \\cdot m^{-3}",
    units_long="Newtons / cubic meter",
    description="Magnitude of force balance error",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["F"],
)
def _Fmag(params, transforms, profiles, data, **kwargs):
    data["|F|"] = safenorm(data["F"], axis=-1)
    return data


@register_compute_fun(
    name="<|F|>_vol",
    label="\\langle |\\mathbf{J} \\times \\mathbf{B} - \\nabla p| \\rangle_{vol}",
    units="N \\cdot m^{-3}",
    units_long="Newtons / cubic meter",
    description="Volume average of magnitude of force balance error",
    dim=0,
    params=[],
    transforms={},
    profiles=[],
    coordinates="",
    data=["|F|", "sqrt(g)", "V"],
    resolution_requirement="rtz",
)
def _Fmag_vol(params, transforms, profiles, data, **kwargs):
    data["<|F|>_vol"] = (
        jnp.sum(data["|F|"] * data["sqrt(g)"] * transforms["grid"].weights) / data["V"]
    )
    return data


@register_compute_fun(
    name="e^helical",
    label="B^{\\theta} \\nabla \\zeta - B^{\\zeta} \\nabla \\theta",
    units="T \\cdot m^{-2}",
    units_long="Tesla / square meter",
    description="Helical basis vector",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B^theta", "B^zeta", "e^theta", "e^zeta"],
)
def _e_sup_helical(params, transforms, profiles, data, **kwargs):
    data["e^helical"] = (
        data["B^zeta"] * data["e^theta"].T - data["B^theta"] * data["e^zeta"].T
    ).T
    return data


@register_compute_fun(
    name="e^helical*sqrt(g)",
    label="\\sqrt{g}(B^{\\theta} \\nabla \\zeta - B^{\\zeta} \\nabla \\theta)",
    units="T \\cdot m^{2}",
    units_long="Tesla * square meter",
    description="Helical basis vector weighted by 3-D volume Jacobian",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["B^theta", "B^zeta", "e^theta*sqrt(g)", "e^zeta", "sqrt(g)"],
)
def _e_sup_helical_times_sqrt_g(params, transforms, profiles, data, **kwargs):
    data["e^helical*sqrt(g)"] = (
        data["B^zeta"] * data["e^theta*sqrt(g)"].T
        - (data["sqrt(g)"] * data["B^theta"]) * data["e^zeta"].T
    ).T
    return data


@register_compute_fun(
    name="|e^helical|",
    label="|B^{\\theta} \\nabla \\zeta - B^{\\zeta} \\nabla \\theta|",
    units="T \\cdot m^{-2}",
    units_long="Tesla / square meter",
    description="Magnitude of helical basis vector",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^helical"],
)
def _e_sup_helical_mag(params, transforms, profiles, data, **kwargs):
    data["|e^helical|"] = jnp.linalg.norm(data["e^helical"], axis=-1)
    return data


@register_compute_fun(
    name="|e^helical*sqrt(g)|",
    label="|\\sqrt{g}(B^{\\theta} \\nabla \\zeta - B^{\\zeta} \\nabla \\theta)|",
    units="T \\cdot m^{2}",
    units_long="Tesla * square meter",
    description="Magnitude of helical basis vector weighted by 3-D volume Jacobian",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["e^helical*sqrt(g)"],
)
def _e_sup_helical_times_sqrt_g_mag(params, transforms, profiles, data, **kwargs):
    data["|e^helical*sqrt(g)|"] = jnp.linalg.norm(data["e^helical*sqrt(g)"], axis=-1)
    return data


@register_compute_fun(
    name="F_anisotropic",
    label="F_{\\mathrm{anisotropic}}",
    units="N \\cdot m^{-3}",
    units_long="Newtons / cubic meter",
    description="Anisotropic force balance error",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["J", "B", "grad(beta_a)", "beta_a", "grad(|B|^2)", "grad(p)"],
)
def _F_anisotropic(params, transforms, profiles, data, **kwargs):
    data["F_anisotropic"] = (
        (1 - data["beta_a"]) * cross(data["J"], data["B"]).T
        - dot(data["B"], data["grad(beta_a)"]) * data["B"].T / mu_0
        - data["beta_a"] * data["grad(|B|^2)"].T / (2 * mu_0)
        - data["grad(p)"].T
    ).T

    return data


@register_compute_fun(
    name="W_B",
    label="W_B",
    units="J",
    units_long="Joules",
    description="Plasma magnetic energy",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["|B|", "sqrt(g)"],
    resolution_requirement="rtz",
)
def _W_B(params, transforms, profiles, data, **kwargs):
    data["W_B"] = jnp.sum(
        data["|B|"] ** 2 * data["sqrt(g)"] * transforms["grid"].weights
    ) / (2 * mu_0)
    return data


@register_compute_fun(
    name="W_Bpol",
    label="W_{B,pol}",
    units="J",
    units_long="Joules",
    description="Plasma magnetic energy in poloidal field",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["B", "sqrt(g)"],
    resolution_requirement="rtz",
)
def _W_Bpol(params, transforms, profiles, data, **kwargs):
    data["W_Bpol"] = jnp.sum(
        dot(data["B"][:, (0, 2)], data["B"][:, (0, 2)])
        * data["sqrt(g)"]
        * transforms["grid"].weights
    ) / (2 * mu_0)
    return data


@register_compute_fun(
    name="W_Btor",
    label="W_{B,tor}",
    units="J",
    units_long="Joules",
    description="Plasma magnetic energy in toroidal field",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["B", "sqrt(g)"],
    resolution_requirement="rtz",
)
def _W_Btor(params, transforms, profiles, data, **kwargs):
    data["W_Btor"] = jnp.sum(
        data["B"][:, 1] ** 2 * data["sqrt(g)"] * transforms["grid"].weights
    ) / (2 * mu_0)
    return data


@register_compute_fun(
    name="W_p",
    label="W_p",
    units="J",
    units_long="Joules",
    description="Plasma thermodynamic energy",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["p", "sqrt(g)"],
    gamma="float: Adiabatic index. Default 0",
    resolution_requirement="rtz",
)
def _W_p(params, transforms, profiles, data, **kwargs):
    data["W_p"] = jnp.sum(data["p"] * data["sqrt(g)"] * transforms["grid"].weights) / (
        kwargs.get("gamma", 0) - 1
    )
    return data


@register_compute_fun(
    name="W",
    label="W",
    units="J",
    units_long="Joules",
    description="Plasma total energy",
    dim=0,
    params=[],
    transforms={},
    profiles=[],
    coordinates="",
    data=["W_B", "W_p"],
)
def _W(params, transforms, profiles, data, **kwargs):
    data["W"] = data["W_B"] + data["W_p"]
    return data


@register_compute_fun(
    name="<beta>_vol",
    label="\\langle \\beta \\rangle_{vol}",
    units="~",
    units_long="None",
    description="Normalized plasma pressure",
    dim=0,
    params=[],
    transforms={},
    profiles=[],
    coordinates="",
    data=["W_p", "W_B"],
)
def _beta_vol(params, transforms, profiles, data, **kwargs):
    data["<beta>_vol"] = jnp.abs(data["W_p"] / data["W_B"])
    return data


@register_compute_fun(
    name="<beta_pol>_vol",
    label="\\langle \\beta_{pol} \\rangle_{vol}",
    units="~",
    units_long="None",
    description="Normalized poloidal plasma pressure",
    dim=0,
    params=[],
    transforms={},
    profiles=[],
    coordinates="",
    data=["W_p", "W_Bpol"],
)
def _beta_volpol(params, transforms, profiles, data, **kwargs):
    data["<beta_pol>_vol"] = jnp.abs(data["W_p"] / data["W_Bpol"])
    return data


@register_compute_fun(
    name="<beta_tor>_vol",
    label="\\langle \\beta_{tor} \\rangle_{vol}",
    units="~",
    units_long="None",
    description="Normalized toroidal plasma pressure",
    dim=0,
    params=[],
    transforms={},
    profiles=[],
    coordinates="",
    data=["W_p", "W_Btor"],
)
def _beta_voltor(params, transforms, profiles, data, **kwargs):
    data["<beta_tor>_vol"] = jnp.abs(data["W_p"] / data["W_Btor"])
    return data


@register_compute_fun(
    name="P_ISS04",
    label="P_{ISS04}",
    units="W",
    units_long="Watts",
    description="Heating power required by the ISS04 energy confinement time scaling",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["a", "iota", "rho", "R0", "W_p", "<ne>_vol", "<|B|>_axis"],
    method="str: Interpolation method. Default 'cubic'.",
    H_ISS04="float: ISS04 confinement enhancement factor. Default 1.",
)
def _P_ISS04(params, transforms, profiles, data, **kwargs):
    rho = transforms["grid"].compress(data["rho"], surface_label="rho")
    iota = transforms["grid"].compress(data["iota"], surface_label="rho")
    fx = {}
    if "iota_r" in data:
        fx["fx"] = transforms["grid"].compress(
            data["iota_r"]
        )  # noqa: unused dependency
    iota_23 = interp1d(2 / 3, rho, iota, method=kwargs.get("method", "cubic"), **fx)
    data["P_ISS04"] = 1e6 * (  # MW -> W
        jnp.abs(data["W_p"] / 1e6)  # J -> MJ
        / (
            0.134
            * data["a"] ** 2.28  # m
            * data["R0"] ** 0.64  # m
            * (data["<ne>_vol"] / 1e19) ** 0.54  # 1/m^3 -> 1e19/m^3
            * data["<|B|>_axis"] ** 0.84  # T
            * iota_23**0.41
            * kwargs.get("H_ISS04", 1)
        )
    ) ** (1 / 0.39)
    return data


@register_compute_fun(
    name="P_fusion",
    label="P_{fusion}",
    units="W",
    units_long="Watts",
    description="Fusion power",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["ni", "<sigma*nu>", "sqrt(g)"],
    resolution_requirement="rtz",
    fuel="str: Fusion fuel, assuming a 50/50 mix. One of {'DT'}. Default is 'DT'.",
)
def _P_fusion(params, transforms, profiles, data, **kwargs):
    energies = {"DT": 3.52e6 + 14.06e6}  # eV
    fuel = kwargs.get("fuel", "DT")
    energy = energies.get(fuel)

    reaction_rate = jnp.sum(
        data["ni"] ** 2
        * data["<sigma*nu>"]
        * data["sqrt(g)"]
        * transforms["grid"].weights
    )  # reactions/s
    data["P_fusion"] = reaction_rate * energy * elementary_charge  # J/s
    return data
